# -*- coding: utf-8 -*-
import logging
import sys
from typing import Optional, Set, Tuple

from nbsafety.data_model.code_cell import cells
from nbsafety.run_mode import ExecutionMode
from nbsafety.singletons import nbs
from test.utils import make_safety_fixture

logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.ERROR)

# Reset dependency graph before each test
# _safety_fixture, run_cell_ = make_safety_fixture(trace_messages_enabled=True)
_safety_fixture, run_cell_ = make_safety_fixture()


def run_cell(
    cell_content: str, cell_id: Optional[int] = None, fresh_are_reactive: bool = False
) -> Tuple[int, Set[int]]:
    orig_mode = nbs().mut_settings.exec_mode
    try:
        if fresh_are_reactive:
            nbs().mut_settings.exec_mode = ExecutionMode.REACTIVE
        executed_cells = set()
        reactive_cells = set()
        next_content_to_run = cell_content
        next_cell_to_run_id = cell_id
        while next_content_to_run is not None:
            executed_cells.add(
                run_cell_(next_content_to_run, cell_id=next_cell_to_run_id)
            )
            if len(executed_cells) == 1:
                cell_id = next(iter(executed_cells))
            next_content_to_run = None
            checker_result = nbs().check_and_link_multiple_cells()
            if fresh_are_reactive:
                reactive_cells |= checker_result.new_fresh_cells
            else:
                reactive_cells |= checker_result.forced_reactive_cells
            for reactive_cell_id in sorted(reactive_cells - executed_cells):
                next_content_to_run = cells().from_id(reactive_cell_id).content
                next_cell_to_run_id = reactive_cell_id
                break
        return cell_id, executed_cells
    finally:
        nbs().mut_settings.exec_mode = orig_mode
        nbs().reactivity_cleanup()


def run_reactively(
    cell_content: str, cell_id: Optional[int] = None
) -> Tuple[int, Set[int]]:
    return run_cell(cell_content, cell_id=cell_id, fresh_are_reactive=True)


def test_mutate_one_list_entry():
    assert run_reactively("lst = [1, 2, 3]")[1] == {1}
    assert run_reactively("logging.info(lst[0])")[1] == {2}
    assert run_reactively("logging.info(lst[1])")[1] == {3}
    assert run_reactively("logging.info(lst[2])")[1] == {4}
    for i in range(3):
        cell_id, cells_run = run_reactively(f"lst[{i}] += 1")
        assert cells_run - {cell_id} == {i + 2}, "got %s" % cells_run
    cell_id, cells_run = run_reactively("lst.append(3)")
    assert cells_run - {cell_id} == set(), "got %s" % cells_run


def test_simple_reactive_var_load():
    assert run_cell("x = 0")[1] == {1}
    assert run_cell("y = $x + 1")[1] == {2}
    assert run_cell("logging.info($y)")[1] == {3}
    assert run_cell("x = 42")[1] == {4, 2, 3}
    cell_id, cells_run = run_cell("y = 99")
    assert cells_run - {cell_id} == {3}


def test_simple_reactive_var_store():
    assert run_cell("x = 0")[1] == {1}
    assert run_cell("y = x + 1")[1] == {2}
    assert run_cell("logging.info(y)")[1] == {3}
    assert run_cell("$x = 42")[1] == {4, 2}
    cell_id, cells_run = run_cell("$y = 99")
    assert cells_run - {cell_id} == {3}, "got %s" % cells_run


def test_simple_blocked_reactive_var_store():
    assert run_cell("x = 0")[1] == {1}
    assert run_cell("$:y = $x + 1")[1] == {2}
    assert run_cell("logging.info($y)")[1] == {3}
    assert run_cell("x = 42")[1] == {4, 2}
    cell_id, cells_run = run_cell("z = 9001")
    assert cells_run - {cell_id} == set(), "got %s" % (cells_run - {cell_id})
    cell_id, cells_run = run_cell("y = 99")
    assert cells_run - {cell_id} == {3}


def test_simple_blocked_reactive_var_load():
    assert run_cell("x = 0")[1] == {1}
    assert run_cell("y = $:x + 1")[1] == {2}
    assert run_cell("logging.info(y)")[1] == {3}
    assert run_cell("$x = 42")[1] == {4}
    cell_id, cells_run = run_cell("$y = 99")
    assert cells_run - {cell_id} == {3}, "got %s" % cells_run


def test_reactive_function_defn():
    assert run_cell("x = 0")[1] == {1}
    assert run_cell("def f(): return x")[1] == {2}
    assert run_cell("logging.info(f())")[1] == {3}
    assert run_cell("def $f(): return x + 3")[1] == {4, 3}


def test_reactive_function_call():
    assert run_cell("x = 0")[1] == {1}
    assert run_cell("def f(): return $x")[1] == {2}
    assert run_cell("logging.info(f())")[1] == {3}
    assert run_cell("x = 42")[1] == {4, 3}


def test_reactive_store_to_global_var_from_function_call():
    assert run_cell("x = 0")[1] == {1}
    assert run_cell("def f(): global x; $x = 42")[1] == {2}
    assert run_cell("logging.info(x)")[1] == {3}
    assert run_cell("f()")[1] == {4, 3}


def test_reactive_store_to_local_var_from_function_call():
    assert run_cell("x = 0")[1] == {1}
    assert run_cell("def f(): $x = 42")[1] == {2}
    assert run_cell("logging.info(x)")[1] == {3}
    assert run_cell("f()")[1] == {4}


if sys.version_info >= (3, 8):

    def test_reactive_attr_load():
        assert (
            run_cell(
                """
            from dataclasses import dataclass
            
            @dataclass
            class Example:
                foo: str
                bar: int
            
            ex = Example('hi', 42)
            """
            )[1]
            == {1}
        )
        assert run_cell("logging.info($ex)")[1] == {2}
        assert run_cell("logging.info(ex.$foo)")[1] == {3}
        assert run_cell("logging.info(ex.$bar)")[1] == {4}
        assert run_cell("logging.info($ex.foo)")[1] == {5}
        assert run_cell("logging.info($ex.bar)")[1] == {6}
        cell_id, cells_run = run_cell('ex.foo = "hello"')
        assert cells_run - {cell_id} == {2, 3, 5, 6}
        cell_id, cells_run = run_cell("ex.bar = 9001")
        assert cells_run - {cell_id} == {2, 4, 5, 6}
        cell_id, cells_run = run_cell('ex = Example("foo", 0)')
        assert cells_run - {cell_id} == {2, 3, 4, 5, 6}

    def test_reactive_attr_store():
        assert (
            run_cell(
                """
            from dataclasses import dataclass
            
            @dataclass
            class Example:
                foo: str
                bar: int
            
            ex = Example('hi', 42)
            """
            )[1]
            == {1}
        )
        assert run_cell("logging.info(ex)")[1] == {2}
        assert run_cell("logging.info(ex.foo)")[1] == {3}
        assert run_cell("logging.info(ex.bar)")[1] == {4}
        assert run_cell("logging.info(ex.$foo)")[1] == {5}
        assert run_cell("logging.info(ex.$bar)")[1] == {6}
        cell_id, cells_run = run_cell('ex.$foo = "hello"')
        assert cells_run - {cell_id} == {2, 3, 5}, "got %s" % (cells_run - {cell_id})
        cell_id, cells_run = run_cell("ex.$bar = 9001")
        assert cells_run - {cell_id} == {2, 4, 6}
        cell_id, cells_run = run_cell('$ex.foo = "wat"')
        assert cells_run - {cell_id} == {2, 3, 4, 5, 6}, "got %s" % (
            cells_run - {cell_id}
        )
        cell_id, cells_run = run_cell('ex = Example("foo", 0)')
        assert cells_run - {cell_id} == {5, 6}
        cell_id, cells_run = run_cell('$ex = Example("foo", 0)')
        assert cells_run - {cell_id} == {2, 3, 4, 5, 6}

    def test_blocked_reactive_attr_store():
        assert (
            run_cell(
                """
            from dataclasses import dataclass
            
            @dataclass
            class Example:
                foo: str
                bar: int
            
            ex = Example('hi', 42)
            """
            )[1]
            == {1}
        )
        assert run_cell("logging.info($ex)")[1] == {2}
        assert run_cell("logging.info(ex.$foo)")[1] == {3}
        assert run_cell("logging.info(ex.$bar)")[1] == {4}
        assert run_cell("logging.info($ex.foo)")[1] == {5}
        assert run_cell("logging.info($ex.bar)")[1] == {6}
        cell_id, cells_run = run_cell('ex.$:foo = "hello"')
        assert cells_run - {cell_id} == set(), "got %s" % (cells_run - {cell_id})
        cell_id, cells_run = run_cell("ex.$:bar = 9001")
        assert cells_run - {cell_id} == set(), "got %s" % (cells_run - {cell_id})
        cell_id, cells_run = run_cell('$:ex = Example("foo", 0)')
        assert cells_run - {cell_id} == set(), "got %s" % (cells_run - {cell_id})

    def test_blocked_reactive_attr_load():
        assert (
            run_cell(
                """
            from dataclasses import dataclass
            
            @dataclass
            class Example:
                foo: str
                bar: int
            
            ex = Example('hi', 42)
            """
            )[1]
            == {1}
        )
        assert run_cell("logging.info(ex)")[1] == {2}
        assert run_cell("logging.info($:ex.foo)")[1] == {3}
        assert run_cell("logging.info($:ex.bar)")[1] == {4}
        assert run_cell("logging.info(ex.$:foo)")[1] == {5}
        assert run_cell("logging.info(ex.$:bar)")[1] == {6}
        cell_id, cells_run = run_cell('ex.$foo = "hello"')
        assert cells_run - {cell_id} == {2}, "got %s" % (cells_run - {cell_id})
        cell_id, cells_run = run_cell("ex.$bar = 9001")
        assert cells_run - {cell_id} == {2}
        cell_id, cells_run = run_cell('$:ex.foo = "wat"')
        assert cells_run - {cell_id} == set(), "got %s" % (cells_run - {cell_id})
        cell_id, cells_run = run_cell('ex = Example("foo", 0)')
        assert cells_run - {cell_id} == set()
        cell_id, cells_run = run_cell('$ex = Example("foo", 0)')
        assert cells_run - {cell_id} == {2, 5, 6}

    def test_store_after_blocked_store_reactively_executes():
        assert run_cell("x = 0")[1] == {1}
        assert run_cell("logging.info(x)")[1] == {2}
        assert run_cell("$:x = 42")[1] == {3}
        assert run_cell("$x = 42")[1] == {4, 2}

    def test_nested_reactive_references():
        assert run_cell("x = 42")[1] == {1}
        assert (
            run_cell(
                """
            def assert_nonzero(v):
                assert v != 0, "Got 0 for v!"
            """
            )[1]
            == {2}
        )
        assert run_cell("$assert_nonzero($x)")[1] == {3}
        assert (
            run_cell(
                """
            def assert_nonzero(v):
                assert v != 0, "v can't be 0"
            """
            )[1]
            == {3, 4}
        )
        rerun = run_cell("x = 43")[1]
        assert rerun == {3, 6}, "got %s" % rerun

    def test_nested_reactive_references_2():
        assert run_cell("x = 42")[1] == {1}
        assert (
            run_cell(
                """
            def assert_nonzero(v):
                assert v != 0, "Got 0 for v!"
            """
            )[1]
            == {2}
        )
        assert run_cell("assert_nonzero($x)")[1] == {3}
        assert (
            run_cell(
                """
            def $assert_nonzero(v):
                assert v != 0, "v can't be 0"
            """
            )[1]
            == {3, 4}
        )
        rerun = run_cell("x = 43")[1]
        assert rerun == {3, 6}, "got %s" % rerun



# -*- coding: utf-8 -*-
import ast

from nbsafety.analysis.symbol_ref import SymbolRef, Atom
from .utils import make_safety_fixture


_safety_fixture, _ = make_safety_fixture()


def at(sym: str, **kwargs) -> Atom:
    return Atom(sym, **kwargs)


def cat(sym: str, **kwargs) -> Atom:
    return at(sym, is_callpoint=True, **kwargs)


def test_unit_chain():
    node = ast.parse("sym").body[0].value
    symchain = SymbolRef(node)
    assert symchain.chain == (at("sym"),)
    node = ast.parse("sym()").body[0].value
    symchain = SymbolRef(node)
    assert symchain.chain == (cat("sym"),)


def test_basic():
    node = ast.parse("a.b.c.d.e.f").body[0].value
    symchain = SymbolRef(node)
    assert symchain.chain == tuple(at(s) for s in ("a", "b", "c", "d", "e", "f"))


def test_calls_none_at_endpoints():
    node = ast.parse("a.b.c().d.e().f").body[0].value
    symchain = SymbolRef(node)
    assert symchain.chain == (at("a"), at("b"), cat("c"), at("d"), cat("e"), at("f"))


def test_calls_at_endpoints():
    node = ast.parse("a().b.c().d.e.f()").body[0].value
    symchain = SymbolRef(node)
    assert symchain.chain == (cat("a"), at("b"), cat("c"), at("d"), at("e"), cat("f"))


def test_hash():
    symchain_set = set()
    symchain_set.add(SymbolRef("f.read()"))
    symchain_set.add(SymbolRef("f.read"))
    symchain_set.add("f")
    symchain_set.add(("f", "read"))
    assert len(symchain_set) == 4



# -*- coding: utf-8 -*-
import argparse
import ast
import astunparse
import logging
import shlex
from typing import cast, Iterable, List, Optional

from nbsafety.data_model.code_cell import cells
from nbsafety.data_model.data_symbol import DataSymbol
from nbsafety.data_model.timestamp import Timestamp
from nbsafety.run_mode import FlowOrder, ExecutionMode, ExecutionSchedule
from nbsafety.singletons import nbs
from nbsafety.tracing.symbol_resolver import resolve_rval_symbols

logger = logging.getLogger(__name__)
logger.setLevel(logging.WARNING)


USAGE = """Options:

[deps|show_deps|show_dependencies] <symbol_1>, <symbol_2> ...: 
    - This will print out the dependencies for given symbols.
      Multiple symbols should be separated with commas.

[stale|show_stale]: 
    - This will print out all the global variables that are stale. 

slice <cell_num>:
    - This will print the code necessary to reconstruct <cell_num> using a dynamic
      program slicing algorithm."""


def show_deps(symbols: str) -> Optional[str]:
    usage = "Usage: %safety show_[deps|dependencies] <symbol_1>[, <symbol_2> ...]"
    if len(symbols) == 0:
        logger.warning(usage)
        return None
    try:
        node = cast(ast.Expr, ast.parse(symbols).body[0]).value
    except SyntaxError:
        logger.warning("Could not find symbol metadata for %s", symbols)
        return None
    if isinstance(node, ast.Tuple):
        unresolved_symbols = node.elts
    else:
        unresolved_symbols = [node]
    statements = []
    for unresolved in unresolved_symbols:
        dsyms = resolve_rval_symbols(unresolved, should_update_usage_info=False)
        if len(dsyms) == 0:
            logger.warning(
                "Could not find symbol metadata for %s",
                astunparse.unparse(unresolved).strip(),
            )
        for dsym in dsyms:
            parents = {par for par in dsym.parents if par.is_user_accessible}
            children = {child for child in dsym.children if child.is_user_accessible}
            dsym_extra_info = f"defined cell: {dsym.defined_cell_num}; last updated cell: {dsym.timestamp.cell_num}"
            if dsym.required_timestamp.is_initialized:
                dsym_extra_info += f"; required: {dsym.required_timestamp.cell_num}"
            statements.append(
                "Symbol {} ({}) is dependent on {} and is a parent of {}".format(
                    dsym.full_namespace_path,
                    dsym_extra_info,
                    parents or "nothing",
                    children or "nothing",
                )
            )
    if len(statements) == 0:
        return None
    else:
        return "\n".join(statements)


def show_stale(line_: str) -> Optional[str]:
    usage = "Usage: %safety show_stale [global|all]"
    line = line_.split()
    if len(line) == 0 or line[0] == "global":
        dsym_sets: Iterable[Iterable[DataSymbol]] = [
            nbs().global_scope.all_data_symbols_this_indentation()
        ]
    elif line[0] == "all":
        dsym_sets = nbs().aliases.values()
    else:
        logger.warning(usage)
        return None
    stale_set = set()
    for dsym_set in dsym_sets:
        for data_sym in dsym_set:
            if data_sym.is_stale and not data_sym.is_anonymous:
                stale_set.add(data_sym)
    if not stale_set:
        return "No symbol has stale dependencies for now!"
    else:
        return "Symbol(s) with stale dependencies are: %s" % stale_set


def trace_messages(line_: str) -> None:
    line = line_.split()
    usage = "Usage: %safety trace_messages [enable|disable]"
    if len(line) != 1:
        logger.warning(usage)
        return
    setting = line[0].lower()
    if setting == "on" or setting.startswith("enable"):
        nbs().trace_messages_enabled = True
    elif setting == "off" or setting.startswith("disable"):
        nbs().trace_messages_enabled = False
    else:
        logger.warning(usage)


def set_highlights(cmd: str, rest: str) -> None:
    usage = "Usage: %safety [hls|nohls]"
    if cmd == "hls":
        nbs().mut_settings.highlights_enabled = True
    elif cmd == "nohls":
        nbs().mut_settings.highlights_enabled = False
    else:
        rest = rest.lower()
        if rest == "on" or rest.startswith("enable"):
            nbs().mut_settings.highlights_enabled = True
        elif rest == "off" or rest.startswith("disable"):
            nbs().mut_settings.highlights_enabled = False
        else:
            logger.warning(usage)


_SLICE_PARSER = argparse.ArgumentParser("slice")
_SLICE_PARSER.add_argument("cell_num", nargs="?", type=int, default=None)
_SLICE_PARSER.add_argument("--stmt", "--stmts", action="store_true")
_SLICE_PARSER.add_argument("--tag", nargs="?", type=str, default=None)


def make_slice(line: str) -> Optional[str]:
    try:
        args = _SLICE_PARSER.parse_args(shlex.split(line))
    except:
        return None
    tag = args.tag
    slice_cells = None
    cell_num = args.cell_num
    if cell_num is None:
        if tag is None:
            cell_num = cells().exec_counter() - 1
    if cell_num is not None:
        slice_cells = {cells().from_timestamp(cell_num)}
    elif args.tag is not None:
        if tag.startswith("$"):
            tag = tag[1:]
            cells().current_cell().mark_as_reactive_for_tag(tag)
        slice_cells = cells().from_tag(tag)
    if slice_cells is None:
        logger.warning("Cell(s) have not yet been run")
    elif len(slice_cells) == 0 and tag is not None:
        logger.warning("No cell(s) for tag: %s", tag)
    else:
        deps = list(
            cells().compute_slice_for_cells(slice_cells, stmt_level=args.stmt).items()
        )
        deps.sort()
        return "\n\n".join(
            f"# Cell {cell_num}\n" + content for cell_num, content in deps
        )
    return None


def set_exec_mode(line_: str) -> None:
    usage = f"Usage: %safety mode [{ExecutionMode.NORMAL}|{ExecutionMode.REACTIVE}]"
    try:
        exec_mode = ExecutionMode(line_.strip())
    except ValueError:
        logger.warning(usage)
        return
    nbs().mut_settings.exec_mode = exec_mode


def set_exec_schedule(line_: str) -> None:
    usage = f"Usage: %safety schedule [{ExecutionSchedule.LIVENESS_BASED}|{ExecutionSchedule.DAG_BASED}|{ExecutionSchedule.STRICT}]"
    if line_.startswith("liveness"):
        schedule = ExecutionSchedule.LIVENESS_BASED
    elif line_.startswith("dag"):
        schedule = ExecutionSchedule.DAG_BASED
    elif line_.startswith("strict"):
        if nbs().mut_settings.flow_order != FlowOrder.IN_ORDER:
            logger.warning(
                "Strict schedule only applicable for forward data flow; skipping"
            )
            return
        schedule = ExecutionSchedule.STRICT
    else:
        logger.warning(usage)
        return
    nbs().mut_settings.exec_schedule = schedule


def set_flow_order(line_: str) -> None:
    line_ = line_.lower().strip()
    usage = f"Usage: %safety flow [{FlowOrder.ANY_ORDER}|{FlowOrder.IN_ORDER}]"
    if line_.startswith("any") or line_ in ("unordered", "both"):
        flow_order = FlowOrder.ANY_ORDER
    elif line_.startswith("in") or line_ in ("ordered", "linear"):
        flow_order = FlowOrder.IN_ORDER
    else:
        logger.warning(usage)
        return
    nbs().mut_settings.flow_order = flow_order



# -*- coding: utf-8 -*-
from enum import Enum
import os


class SafetyRunMode(Enum):
    PRODUCTION = "NBSAFETY_PRODUCTION"
    DEVELOP = "NBSAFETY_DEVELOP"

    @staticmethod
    def get():
        if (
            SafetyRunMode.DEVELOP.value in os.environ
            and str(os.environ[SafetyRunMode.DEVELOP.value]) == "1"
        ):
            return SafetyRunMode.DEVELOP
        else:
            return SafetyRunMode.PRODUCTION


class ExecutionMode(Enum):
    NORMAL = "normal"
    REACTIVE = "reactive"


class ExecutionSchedule(Enum):
    LIVENESS_BASED = "liveness_based"
    DAG_BASED = "dag_based"
    STRICT = "strict"


class FlowOrder(Enum):
    ANY_ORDER = "any_order"
    IN_ORDER = "in_order"



# -*- coding: utf-8 -*-
import functools
import logging

from nbsafety.data_model.code_cell import cells
from nbsafety.singletons import nbs
from test.utils import make_safety_fixture

logging.basicConfig(level=logging.ERROR)

# Reset dependency graph before each test
# _safety_fixture, run_cell_ = make_safety_fixture(trace_messages_enabled=True)
_safety_fixture, run_cell_ = make_safety_fixture()


def dynamic_or_static_only_test(test_f, dynamic_enabled, static_enabled):
    @functools.wraps(test_f)
    def dynamic_test_f(*args, **kwargs):
        orig_dynamic_enabled = nbs().mut_settings.dynamic_slicing_enabled
        orig_static_enabled = nbs().mut_settings.dynamic_slicing_enabled
        try:
            nbs().mut_settings.dynamic_slicing_enabled = dynamic_enabled
            nbs().mut_settings.static_slicing_enabled = static_enabled
            test_f(*args, **kwargs)
        finally:
            nbs().mut_settings.dynamic_slicing_enabled = orig_dynamic_enabled
            nbs().mut_settings.static_slicing_enabled = orig_static_enabled

    return dynamic_test_f


def dynamic_only_test(test_f):
    return dynamic_or_static_only_test(test_f, True, False)


def static_only_test(test_f):
    return dynamic_or_static_only_test(test_f, False, True)


def run_cell(cell):
    # print()
    # print('*******************************************')
    # print('running', cell)
    # print('*******************************************')
    # print()
    run_cell_(cell)


def compute_slice_stmts(cell_num):
    return cells().from_timestamp(cell_num).compute_slice_stmts()


def compute_slice(cell_num):
    return cells().from_timestamp(cell_num).compute_slice()


def num_stmts_in_slice(cell_num: int) -> int:
    return sum(len(stmts) for stmts in compute_slice_stmts(cell_num).values())


def test_simple():
    run_cell("a = 1")
    run_cell("b = 2")
    run_cell("logging.info(a)")
    run_cell("c = a + b")
    deps = set(compute_slice(4).keys())
    assert deps == {1, 2, 4}, "got %s" % deps


def test_dynamic_symbol_usage():
    run_cell("x = 5")
    run_cell(
        """
        class Foo:
            def foo(self):
                return x
        """
    )
    run_cell(
        """
        def foo():
            return Foo()
        """
    )
    run_cell("logging.info(foo().foo())")
    deps = set(compute_slice(4).keys())
    assert deps == {1, 2, 3, 4}, "got %s" % deps
    slice_size = num_stmts_in_slice(4)
    assert slice_size == 4, "got %d" % slice_size


def test_nested_symbol_usage():
    run_cell("lst = [1, 2, 3, 4, 5]")
    run_cell("lst[1] = 3")
    run_cell("logging.info(lst[1])")
    deps = set(compute_slice(3).keys())
    assert deps == {1, 2, 3}, "got %s" % deps
    slice_size = num_stmts_in_slice(3)
    assert slice_size == 3, "got %d" % slice_size


def test_nested_symbol_usage_with_variable_subscript():
    for static_slicing_enabled in [True, False]:
        if (
            not static_slicing_enabled
            and not nbs().mut_settings.dynamic_slicing_enabled
        ):
            continue
        orig_enabled = nbs().mut_settings.static_slicing_enabled
        try:
            nbs().mut_settings.static_slicing_enabled = static_slicing_enabled
            run_cell("x = 1")
            run_cell("lst = [1, 2, 3, 4, 5]")
            run_cell("lst[x] = 3")
            run_cell("logging.info(lst[1])")
            deps = set(compute_slice(4).keys())
            assert deps == {1, 2, 3, 4}, "got %s" % deps
            slice_size = num_stmts_in_slice(4)
            assert slice_size == 4, "got %d" % slice_size
        finally:
            nbs().mut_settings.static_slicing_enabled = orig_enabled


@static_only_test
def test_liveness_timestamps():
    run_cell(
        """
        x = 0
        if True:
            y = 1
        else:
            y = 2
        """
    )
    run_cell("z = 42")
    run_cell("logging.info(x + 1)")
    deps = set(compute_slice(3).keys())
    assert deps == {1, 3}, "got %s" % deps
    slice_size = num_stmts_in_slice(3)
    assert slice_size == 2, "got %d" % slice_size


def test_list_mutations():
    run_cell("dummy = 0")
    run_cell("lst = []")
    run_cell("lst.append(1)")
    run_cell("lst.append(2)")
    run_cell("lst.append(3); lst.append(4)")
    run_cell("logging.info(lst)")
    deps = set(compute_slice(5).keys())
    assert deps == {2, 3, 4, 5}, "got %s" % deps
    slice_size = num_stmts_in_slice(5)
    assert slice_size == 5, "got %d" % slice_size


def test_imports():
    run_cell("import numpy as np")
    run_cell("arr = np.zeros((5,))")
    run_cell("logging.info(arr * 3)")
    deps = set(compute_slice(3).keys())
    assert deps == {1, 2, 3}, "got %s" % deps
    slice_size = num_stmts_in_slice(3)
    assert slice_size == 3, "got %d" % slice_size


def test_handle_stale():
    run_cell("a = 1")
    run_cell("b = 2 * a")
    run_cell("a = 2")
    run_cell("logging.info(b)")
    run_cell("logging.info(b)")
    deps = set(compute_slice(4).keys())
    assert deps == {1, 2, 4}, "got %s" % deps
    slice_size = num_stmts_in_slice(4)
    assert slice_size == 3, "got %d" % slice_size


@dynamic_only_test
def test_multiple_versions_captured():
    run_cell("x = 0")
    run_cell("logging.info(x); y = 7")
    run_cell("x = 5")
    run_cell("logging.info(x + y)")
    deps = set(compute_slice(4).keys())
    assert deps == {1, 2, 3, 4}, "got %s" % deps
    slice_size = num_stmts_in_slice(4)
    assert slice_size == 3, "got %d" % slice_size


def test_version_used_when_live():
    run_cell("x = 0")
    run_cell(
        """
        if True:
            y = 7
        else:
            # even though this branch is not taken,
            # liveness-based usage should detect the
            # version of `x` used at the time it was
            # live, meaning cell 1 should get included
            # in the slice
            logging.info(x)
        """
    )
    run_cell("x = 5")
    run_cell("logging.info(x + y)")
    deps = set(compute_slice(4).keys())
    assert deps == {1, 2, 3, 4}, "got %s" % deps
    slice_size = num_stmts_in_slice(4)
    assert slice_size == 4, "got %d" % slice_size


def test_no_definitely_spurious_cells():
    run_cell("x = 0")
    run_cell(
        """
        if True:
            y = 7
        else:
            # even though this branch is not taken,
            # liveness-based usage should detect the
            # version of `x` used at the time it was
            # live, meaning cell 1 should get included
            # in the slice
            logging.info(x)
        """
    )
    run_cell("x = 5")
    run_cell("logging.info(y)")
    deps = set(compute_slice(4).keys())
    assert deps == {1, 2, 4}, "got %s" % deps
    slice_size = num_stmts_in_slice(4)
    assert slice_size == 3, "got %d" % slice_size


def test_parent_usage_includes_child_update():
    run_cell("lst = [3]")
    run_cell("lst[0] += 1")
    run_cell("lst2 = lst + [5]")
    deps = set(compute_slice(3).keys())
    assert deps == {1, 2, 3}, "got %s" % deps
    slice_size = num_stmts_in_slice(3)
    assert slice_size == 3, "got %d" % slice_size


def test_object_subscripting():
    run_cell(
        """
        class Foo:
            def __init__(self):
                self.counter = 0
                
            def inc(self):
                self.counter += 1

            def bar(self):
                return {'baz': 0}
        """
    )
    run_cell("obj = Foo()")
    run_cell("name = 'baz'")
    run_cell("obj.bar()[name]")
    run_cell("something_i_dont_care_about = obj.bar()[name]")
    run_cell("something_i_care_about = obj.bar()[name]")
    deps = set(compute_slice(6).keys())
    assert deps == {1, 2, 3, 6}, "got %s" % deps
    slice_size = num_stmts_in_slice(6)
    assert slice_size == 4, "got %d" % slice_size


def test_complicated_subscripting():
    run_cell(
        """
        class Foo:
            def __init__(self):
                self.counter = 0
                
            def inc(self):
                self.counter += 1

            def bar(self, indicator):
                if indicator == 1:
                    return the_dictionary
                else:
                    return x + 5
            
            def new(self, indicator):
                if indicator == 1:
                    return Foo()
                
                return Bar()
        """
    )
    run_cell(
        """
        class Bar:
            def __init__(self):
                self.counter = 0
            
            def foo(self, indicator):
                if indicator == 1:
                    return the_dictionary
                
                return Foo()
        """
    )
    run_cell("x = 1")
    run_cell("the_dictionary = {'something': 1}")
    run_cell("Foo().new(0).foo(1)")
    deps = set(compute_slice(5).keys())
    assert deps == {1, 2, 4, 5}, "got %s" % deps
    slice_size = num_stmts_in_slice(5)
    assert slice_size == 4, "got %d" % slice_size


def test_complicated_subscripting_use_conditional():
    run_cell(
        """
        class Foo:
            def __init__(self):
                self.counter = 0
                
            def inc(self):
                self.counter += 1

            def bar(self, indicator):
                if indicator == 1:
                    return the_dictionary
                else:
                    return x + 5
            
            def new(self, indicator):
                if indicator == 1:
                    return Foo()
                
                return Bar()
        """
    )
    run_cell(
        """
        class Bar:
            def __init__(self):
                self.counter = 0
            
            def foo(self, indicator):
                if indicator == 1:
                    return the_dictionary
                
                return Foo()
        """
    )
    run_cell("x = 1")
    run_cell("the_dictionary = {'something': 1}")
    run_cell(
        """
        y = Foo().new(0).foo(1)
        z = Foo().new(1).bar(0)
        """
    )
    run_cell("logging.info(z)")
    run_cell("logging.info(y)")
    deps = set(compute_slice(6).keys())
    assert deps == {1, 2, 3, 4, 5, 6}, "got %s" % deps
    slice_size = num_stmts_in_slice(6)
    assert slice_size == 4, "got %d" % slice_size
    slice_size = num_stmts_in_slice(7)
    assert slice_size == 5, "got %d" % slice_size


def test_non_relevant_child_symbol_modified():
    run_cell("lst = [0, 1, 2]")
    run_cell("lst[0] += 1")
    run_cell("lst[0] += 1")
    run_cell("lst[1] += 1")
    run_cell("lst[0] += 1")
    run_cell("lst[0] += 1")
    run_cell("logging.info(lst[1])")
    run_cell("lst[1] = 42")
    deps = set(compute_slice(7).keys())
    assert deps == {1, 4, 7}, "got %s" % deps
    deps = set(compute_slice(8).keys())
    assert deps == {1, 8}, "got %s" % deps
    slice_size = num_stmts_in_slice(7)
    assert slice_size == 3, "got %d" % slice_size


@dynamic_only_test
def test_dynamic_only_increment():
    nbs().mut_settings.static_slicing_enabled = False
    run_cell("x = 0")
    run_cell("x += 1")
    run_cell("logging.info(x)")
    deps = set(compute_slice(3).keys())
    assert deps == {1, 2, 3}, "got %s" % deps
    slice_size = num_stmts_in_slice(3)
    assert slice_size == 3, "got %d" % slice_size


@dynamic_only_test
def test_dynamic_only_variable_subscript():
    nbs().mut_settings.static_slicing_enabled = False
    run_cell("lst = [0, 1, 2]")
    run_cell("x = 0")
    run_cell("lst[x] += 1")
    run_cell("lst[x] += 1")
    run_cell("x += 1")
    run_cell("lst[x] += 1")
    run_cell("x -= 1")
    run_cell("lst[x] += 1")
    run_cell("lst[x] += 1")
    run_cell("x += 1")
    run_cell("logging.info(lst[x])")
    deps = set(compute_slice(11).keys())
    assert deps == {1, 2, 5, 6, 7, 10, 11}, "got %s" % deps
    slice_size = num_stmts_in_slice(11)
    assert slice_size == len(deps), "got %d" % slice_size


@dynamic_only_test
def test_handler():
    run_cell(
        """
        try:
            r = map()
        except:
            success = False
        """
    )
    run_cell(
        """
        try:
            logging.info("%d %d", (1, 2))
        except:
            success = False
        """
    )
    deps = set(compute_slice(2).keys())
    assert deps == {2}, "got %s" % deps
    slice_size = num_stmts_in_slice(2)
    assert slice_size == 1, "got %d" % slice_size


@dynamic_only_test
def test_list_delete():
    run_cell("lst = [0, 1, 2, 3, 4, 5, 6]")
    run_cell("lst.append(7)")
    run_cell("del lst[0]")
    run_cell("del lst[2]")
    run_cell("logging.info(lst[1])")
    deps = set(compute_slice(5).keys())
    assert deps == {1, 2, 3, 5}, "got %s" % deps
    slice_size = num_stmts_in_slice(5)
    assert slice_size == len(deps), "got %d" % slice_size


def test_anonymous_symbols_attached_on_fun_return_do_not_interfere():
    run_cell("y = 3")
    run_cell("def f(): return y + 5")
    run_cell("f()")
    run_cell("sink = f()")
    assert num_stmts_in_slice(4) == 3



# -*- coding: utf-8 -*-
from nbsafety.data_model.annotation_utils import (
    get_type_annotation,
    make_annotation_string,
)


def _make_annotation_str_for_obj(obj):
    return make_annotation_string(get_type_annotation(obj))


def test_dict():
    empty = _make_annotation_str_for_obj({})
    assert empty == "Dict", "got %s" % empty
    assert _make_annotation_str_for_obj({1: 2}) == "Dict[int, int]"
    assert _make_annotation_str_for_obj({1: 2, 2: 3.0}) == "Dict[int, float]"
    ann_str = _make_annotation_str_for_obj({1: 2, 2: 3.0, "foo": "bar"})
    assert ann_str == "Dict[Union[int, str], Union[float, str]]", "got %s" % ann_str
    assert (
        _make_annotation_str_for_obj({1: 2, 2: 3.0, 3: None})
        == "Dict[int, Optional[float]]"
    )


def test_list():
    assert _make_annotation_str_for_obj([]) == "List"
    assert _make_annotation_str_for_obj([3, 6.0, 9]) == "List[float]"
    assert _make_annotation_str_for_obj([3, 9]) == "List[int]"
    ann_str = _make_annotation_str_for_obj([3, None])
    assert ann_str == "List[Optional[int]]", "got %s" % ann_str
    assert _make_annotation_str_for_obj([3, None, 42.0]) == "List[Optional[float]]"
    ann_str = _make_annotation_str_for_obj([3, None, 42.0, "foo"])
    assert ann_str == "List[Union[None, float, str]]", "got %s" % ann_str


def test_set():
    assert _make_annotation_str_for_obj(set()) == "Set"
    assert _make_annotation_str_for_obj({3, 6.0, 9}) == "Set[float]"
    assert _make_annotation_str_for_obj({3, 9}) == "Set[int]"
    assert _make_annotation_str_for_obj({3, None}) == "Set[Optional[int]]"
    assert _make_annotation_str_for_obj({3, None, 42.0}) == "Set[Optional[float]]"
    assert (
        _make_annotation_str_for_obj({3, None, 42.0, "foo"})
        == "Set[Union[None, float, str]]"
    )


def test_tuple():
    assert _make_annotation_str_for_obj(()) == "Tuple"
    assert _make_annotation_str_for_obj((3, 4)) == "Tuple[int, int]"
    assert _make_annotation_str_for_obj((3, 4, 5.0)) == "Tuple[int, int, float]"
    assert _make_annotation_str_for_obj((3, 4, 5, 6)) == "Tuple[int, ...]"
    assert _make_annotation_str_for_obj((3, 4, 5, 6.0)) == "Tuple[float, ...]"


class Foo:
    pass


def test_class():
    ann_str = _make_annotation_str_for_obj(Foo)
    assert ann_str == "Type[{prefix}.Foo]".format(prefix=__name__), "got %s" % ann_str
    assert _make_annotation_str_for_obj(
        (Foo, Foo())
    ) == "Tuple[Type[{prefix}.Foo], {prefix}.Foo]".format(prefix=__name__)
    assert _make_annotation_str_for_obj(
        [Foo, Foo()]
    ) == "List[Union[Type[{prefix}.Foo], {prefix}.Foo]]".format(prefix=__name__)



# -*- coding: utf-8 -*-
import ast
import textwrap

from pyccolo.expr_rewriter import ExprRewriter
from pyccolo.trace_events import TraceEvent
from nbsafety.utils import KeyDict


PROGRAM = textwrap.dedent(
    """
    for i in [foo(x) for x in [1, 2, 3]]:
        try:
            logging.info(i)
        except:
            logging.warning("warning!")
        finally:
            sys.exit(0)
            
    if False:
        asdf(qwer(1, 2, a().b[c,d,e](f.g()).h))[a, 7] = foo.bar
    """
)


def test_ast_rewrite():
    """
    No asserts; just make sure we don't throw an error.
    """
    rewriter = ExprRewriter(
        KeyDict(), {event: (lambda nd: True) for event in TraceEvent}, set()
    )
    assert rewriter.visit(ast.parse(PROGRAM)) is not None



# -*- coding: utf-8 -*-
import logging

from nbsafety.data_model.code_cell import cells
from test.utils import make_safety_fixture

logging.basicConfig(level=logging.ERROR)

# Reset dependency graph before each test
# _safety_fixture, run_cell_ = make_safety_fixture(trace_messages_enabled=True)
_safety_fixture, run_cell = make_safety_fixture()


def test_simple():
    run_cell("x = 0")
    run_cell("y = x + 1")
    assert cells().from_id(2).dynamic_parent_ids == {1}
    assert cells().from_id(1).dynamic_children_ids == {2}, (
        "got %s" % cells().from_id(1)._dynamic_children
    )
    run_cell("z = x + y + 2")
    assert cells().from_id(3).dynamic_parent_ids == {1, 2}
    assert cells().from_id(1).dynamic_children_ids == {2, 3}
    assert cells().from_id(2).dynamic_children_ids == {3}
    run_cell("x = 42")
    assert cells().from_id(3).dynamic_parent_ids == {1, 2}
    assert cells().from_id(1).dynamic_children_ids == {2, 3}
    assert cells().from_id(2).dynamic_children_ids == {3}
    run_cell("y = x + 1")
    assert cells().from_id(3).dynamic_parent_ids == {1, 2}
    assert cells().from_id(1).dynamic_children_ids == {2, 3}
    assert cells().from_id(2).dynamic_children_ids == {3}
    assert cells().from_id(5).dynamic_parent_ids == {4}



# -*- coding: utf-8 -*-
import ast
import logging
import sys
from typing import cast, TYPE_CHECKING, Iterable, List, Optional, Set, Tuple, Union

from nbsafety.analysis.resolved_symbols import ResolvedDataSymbol
from nbsafety.analysis.symbol_ref import LiveSymbolRef, SymbolRef, Atom
from nbsafety.analysis.mixins import (
    SaveOffAttributesMixin,
    SkipUnboundArgsMixin,
    VisitListsMixin,
)
from nbsafety.data_model.timestamp import Timestamp
from nbsafety.run_mode import FlowOrder
from nbsafety.singletons import nbs, tracer

if TYPE_CHECKING:
    from nbsafety.data_model.data_symbol import DataSymbol
    from nbsafety.data_model.scope import Scope

logger = logging.getLogger(__name__)
logger.setLevel(logging.ERROR)


# TODO: have the logger warnings additionally raise exceptions for tests
class ComputeLiveSymbolRefs(
    SaveOffAttributesMixin, SkipUnboundArgsMixin, VisitListsMixin, ast.NodeVisitor
):
    def __init__(
        self, scope: Optional["Scope"] = None, init_killed: Optional[Set[str]] = None
    ) -> None:
        self._scope = scope
        self._module_stmt_counter = 0
        # live symbols also include the stmt counter of when they were live, for slicing purposes later
        self.live: Set[LiveSymbolRef] = set()
        if init_killed is None:
            self.dead: Set[SymbolRef] = set()
        else:
            self.dead = cast("Set[SymbolRef]", init_killed)
        # TODO: use the ast context instead of hacking our own (e.g. ast.Load(), ast.Store(), etc.)
        self._in_kill_context = False
        self._inside_attrsub = False
        self._skip_simple_names = False

    def __call__(self, node: ast.AST) -> Tuple[Set[LiveSymbolRef], Set[SymbolRef]]:
        """
        This function should be called when we want to do a liveness check on a
        cell's corresponding ast.Module.
        """
        # TODO: this will break if we ref a variable in a loop before killing it in the
        #   same loop, since we will add everything on the LHS of an assignment to the killed
        #   set before checking the loop body for live variables
        self.visit(node)
        return self.live, self.dead

    def kill_context(self):
        return self.push_attributes(_in_kill_context=True)

    def live_context(self):
        return self.push_attributes(_in_kill_context=False)

    def attrsub_context(self, inside=True):
        return self.push_attributes(_inside_attrsub=inside, _skip_simple_names=inside)

    def args_context(self):
        return self.push_attributes(_skip_simple_names=False)

    def _add_attrsub_to_live_if_eligible(self, ref: SymbolRef) -> None:
        if ref.nonreactive() in self.dead:
            return
        if len(ref.chain) == 0:
            # can happen if user made syntax error like [1, 2, 3][4, 5, 6] (e.g. forgot comma)
            return
        leading_atom = ref.chain[0]
        if isinstance(leading_atom.value, str):
            if (
                SymbolRef(leading_atom) in self.dead
                or SymbolRef(Atom(leading_atom.value, is_callpoint=False)) in self.dead
            ):
                return
        self.live.add(LiveSymbolRef(ref, self._module_stmt_counter))

    # the idea behind this one is that we don't treat a symbol as dead
    # if it is used on the RHS of an assignment
    def visit_Assign_impl(self, targets, value, aug_assign_target=None) -> None:
        this_assign_live: Set[LiveSymbolRef] = set()
        # we won't mutate overall dead for visiting simple targets, and we need it to avoid adding false positive lives
        with self.push_attributes(live=this_assign_live):
            self.visit(value)
            if aug_assign_target is not None:
                self.visit(aug_assign_target)
        # make a copy, then track the new dead
        this_assign_dead = set(self.dead)
        with self.push_attributes(dead=this_assign_dead):
            with self.kill_context():
                for target in targets:
                    self.visit_Assign_target(target)
        this_assign_dead -= self.dead
        # TODO: ideally under the current abstraction we should
        #  not be resolving static references to symbols here
        if (
            nbs().mut_settings.flow_order == FlowOrder.ANY_ORDER
            and self._scope is not None
            and len(this_assign_live) == 1
            and len(this_assign_dead) == 1
            and not (this_assign_dead <= self.dead)
            and aug_assign_target is None
            and isinstance(value, (ast.Attribute, ast.Subscript, ast.Name))
        ):
            lhs, rhs = [
                get_symbols_for_references(x, self._scope)[0]
                for x in (this_assign_dead, (live.ref for live in this_assign_live))
            ]
            if len(lhs) == 1 and len(rhs) == 1:
                syms: List[DataSymbol] = [next(iter(x)) for x in (lhs, rhs)]
                lhs_sym, rhs_sym = syms[0], syms[1]
                # hack to avoid marking `b` as live when objects are same,
                # or when it was detected that rhs symbol wasn't actually updated
                if (
                    lhs_sym.obj is rhs_sym.obj
                    or lhs_sym.timestamp_excluding_ns_descendents > rhs_sym.timestamp
                ):
                    # either (a) it's a no-op (so treat it as such), or
                    #        (b) lhs is newer and it doesn't make sense to refresh
                    this_assign_live.clear()
        this_assign_dead -= {live.ref for live in this_assign_live}
        # for ref in this_assign_dead:
        #     if isinstance(ref, AttrSubSymbolChain) and len(ref.symbols) > 1:
        #         this_assign_live.add(
        #             (AttrSubSymbolChain(list(ref.symbols[:-1]) + [
        #                 # FIXME: hack to ensure it can't be resolved all the way, so that we use
        #                 #  timestamp_excluding_ns_children instead of timestamp
        #                 CallPoint('<dummy>')
        #             ]), self._module_stmt_counter)
        #         )
        self.live |= this_assign_live
        self.dead |= this_assign_dead

    if sys.version_info >= (3, 8):

        def visit_NamedExpr(self, node: ast.NamedExpr) -> None:
            self.visit_Assign_impl([node.target], node.value)

    def visit_Assign(self, node: ast.Assign) -> None:
        self.visit_Assign_impl(node.targets, node.value)

    def visit_AnnAssign(self, node: ast.AnnAssign) -> None:
        self.visit_Assign_impl([node.target], node.value)

    def visit_AugAssign(self, node: ast.AugAssign) -> None:
        self.visit_Assign_impl([], node.value, aug_assign_target=node.target)

    def visit_Assign_target(
        self,
        target_node: Union[
            ast.Attribute, ast.Name, ast.Subscript, ast.Tuple, ast.List, ast.expr
        ],
    ) -> None:
        if isinstance(target_node, (ast.Name, ast.Attribute, ast.Subscript)):
            self.dead.add(SymbolRef(target_node))
            if isinstance(target_node, ast.Subscript):
                with self.live_context():
                    self.visit(target_node.slice)
        elif isinstance(target_node, (ast.Tuple, ast.List)):
            for elt in target_node.elts:
                self.visit_Assign_target(elt)
        elif isinstance(target_node, ast.Starred):
            self.visit_Assign_target(target_node.value)
        else:
            logger.warning("unsupported type for node %s" % target_node)

    def visit_FunctionDef(self, node: ast.FunctionDef) -> None:
        self.generic_visit(node.args.defaults)
        self.generic_visit(node.decorator_list)
        self.dead.add(SymbolRef(node))

    def visit_Name(self, node: ast.Name) -> None:
        ref = SymbolRef(node)
        if self._in_kill_context:
            self.dead.add(ref)
        elif not self._skip_simple_names and ref not in self.dead:
            if id(node) in tracer().reactive_node_ids:
                ref.chain[0].is_reactive = True
            self.live.add(LiveSymbolRef(ref, self._module_stmt_counter))

    def visit_Tuple_or_List(self, node: Union[ast.List, ast.Tuple]) -> None:
        for elt in node.elts:
            self.visit(elt)

    def visit_List(self, node: ast.List) -> None:
        self.visit_Tuple_or_List(node)

    def visit_Tuple(self, node: ast.Tuple) -> None:
        self.visit_Tuple_or_List(node)

    def visit_For(self, node: ast.For) -> None:
        # Case "for a,b in something: "
        self.visit(node.iter)
        with self.kill_context():
            self.visit(node.target)
        for line in node.body:
            self.visit(line)

    def visit_ClassDef(self, node: ast.ClassDef) -> None:
        self.generic_visit(node.bases)
        self.generic_visit(node.decorator_list)
        self.generic_visit(node.body)
        self.dead.add(SymbolRef(node))

    def visit_Call(self, node: ast.Call) -> None:
        with self.args_context():
            self.generic_visit(node.args)
            for kwarg in node.keywords:
                self.visit(kwarg.value)
        if not self._inside_attrsub:
            self._add_attrsub_to_live_if_eligible(SymbolRef(node))
        with self.attrsub_context():
            self.visit(node.func)

    def visit_Attribute(self, node: ast.Attribute) -> None:
        if not self._inside_attrsub:
            self._add_attrsub_to_live_if_eligible(SymbolRef(node))
        with self.attrsub_context():
            self.visit(node.value)

    def visit_Subscript(self, node: ast.Subscript) -> None:
        if not self._inside_attrsub:
            self._add_attrsub_to_live_if_eligible(SymbolRef(node))
        with self.attrsub_context():
            self.visit(node.value)
        with self.attrsub_context(inside=False):
            self.visit(node.slice)

    def visit_Delete(self, node: ast.Delete) -> None:
        pass

    def visit_GeneratorExp(self, node) -> None:
        self.visit_GeneratorExp_or_DictComp_or_ListComp_or_SetComp(node)

    def visit_DictComp(self, node) -> None:
        self.visit_GeneratorExp_or_DictComp_or_ListComp_or_SetComp(node)

    def visit_ListComp(self, node) -> None:
        self.visit_GeneratorExp_or_DictComp_or_ListComp_or_SetComp(node)

    def visit_SetComp(self, node) -> None:
        self.visit_GeneratorExp_or_DictComp_or_ListComp_or_SetComp(node)

    def visit_GeneratorExp_or_DictComp_or_ListComp_or_SetComp(self, node) -> None:
        # TODO: as w/ for loop, this will have false positives on later live references
        for gen in node.generators:
            self.visit(gen.iter)
            with self.kill_context():
                self.visit(gen.target)
        # visit the elt at the end to ensure we don't add it to live vars if it was one of the generator targets
        self.visit(node.elt)

    def visit_Lambda(self, node: ast.Lambda) -> None:
        with self.kill_context():
            self.visit(node.args)

    def visit_arg(self, node) -> None:
        ref = SymbolRef(node.arg)
        if self._in_kill_context:
            self.dead.add(ref)
        elif not self._skip_simple_names and ref not in self.dead:
            self.live.add(LiveSymbolRef(ref, self._module_stmt_counter))

    def visit_Module(self, node: ast.Module) -> None:
        for child in node.body:
            assert isinstance(child, ast.stmt)
            self.visit(child)
            self._module_stmt_counter += 1


def get_symbols_for_references(
    symbol_refs: Iterable[SymbolRef],
    scope: "Scope",
) -> Tuple[Set["DataSymbol"], Set["DataSymbol"]]:
    dsyms: Set[DataSymbol] = set()
    called_dsyms: Set["DataSymbol"] = set()
    for symbol_ref in symbol_refs:
        for resolved in symbol_ref.gen_resolved_symbols(
            scope, only_yield_final_symbol=True
        ):
            if resolved.is_called:
                called_dsyms.add(resolved.dsym)
            else:
                dsyms.add(resolved.dsym)
    return dsyms, called_dsyms


def get_live_symbols_and_cells_for_references(
    symbol_refs: Set[LiveSymbolRef],
    scope: "Scope",
    cell_ctr: int,
    update_liveness_time_versions: bool = False,
) -> Tuple[Set[ResolvedDataSymbol], Set[int]]:
    live_symbols: Set[ResolvedDataSymbol] = set()
    called_dsyms: Set[Tuple["DataSymbol", int]] = set()
    for live_symbol_ref in symbol_refs:
        for resolved in live_symbol_ref.gen_resolved_symbols(
            scope, only_yield_final_symbol=False, yield_all_intermediate_symbols=True
        ):
            if update_liveness_time_versions:
                ts_to_use = (
                    resolved.dsym.timestamp
                    if resolved.is_last
                    else resolved.dsym.timestamp_excluding_ns_descendents
                )
                liveness_time = Timestamp(cell_ctr, resolved.liveness_timestamp)
                assert liveness_time > ts_to_use
                if ts_to_use.is_initialized:
                    nbs().add_static_data_dep(liveness_time, ts_to_use)
                    resolved.dsym.timestamp_by_liveness_time[liveness_time] = ts_to_use
            if resolved.is_called:
                called_dsyms.add((resolved.dsym, live_symbol_ref.timestamp))
            if not resolved.is_unsafe:
                live_symbols.add(resolved)
    live_from_calls, live_cells = _compute_call_chain_live_symbols_and_cells(
        called_dsyms, cell_ctr, update_liveness_time_versions
    )
    live_symbols |= live_from_calls
    return live_symbols, live_cells


def _compute_call_chain_live_symbols_and_cells(
    live_with_stmt_ctr: Set[Tuple["DataSymbol", int]],
    cell_ctr: int,
    update_liveness_time_versions: bool,
) -> Tuple[Set[ResolvedDataSymbol], Set[int]]:
    seen = set()
    worklist = list(live_with_stmt_ctr)
    live: Set[ResolvedDataSymbol] = set()
    while len(worklist) > 0:
        workitem = worklist.pop()
        if workitem in seen:
            continue
        called_dsym, stmt_ctr = workitem
        # TODO: handle callable classes
        if not called_dsym.is_function:
            continue
        seen.add(workitem)
        live_refs, _ = compute_live_dead_symbol_refs(
            cast(ast.FunctionDef, called_dsym.stmt_node).body,
            init_killed=set(called_dsym.get_definition_args()),
        )
        used_time = Timestamp(cell_ctr, stmt_ctr)
        for symbol_ref in live_refs:
            for resolved in symbol_ref.gen_resolved_symbols(
                called_dsym.call_scope, only_yield_final_symbol=False
            ):
                if resolved.is_called:
                    worklist.append((resolved.dsym, stmt_ctr))
                if resolved.dsym.is_globally_accessible:
                    if not resolved.is_unsafe:
                        live.add(resolved)
                    if update_liveness_time_versions:
                        ts_to_use = (
                            resolved.dsym.timestamp
                            if resolved.is_last
                            else resolved.dsym.timestamp_excluding_ns_descendents
                        )
                        resolved.dsym.timestamp_by_liveness_time[used_time] = ts_to_use
    return live, {called_dsym.timestamp.cell_num for called_dsym, _ in seen}


def compute_live_dead_symbol_refs(
    code: Union[ast.AST, List[ast.stmt], str],
    scope: "Scope" = None,
    init_killed: Optional[Set[str]] = None,
) -> Tuple[Set[LiveSymbolRef], Set[SymbolRef]]:
    if init_killed is None:
        init_killed = set()
    if isinstance(code, str):
        code = ast.parse(code)
    elif isinstance(code, list):
        code = ast.Module(code)
    return ComputeLiveSymbolRefs(scope=scope, init_killed=init_killed)(code)



# -*- coding: utf-8 -*-
import logging

from .utils import make_safety_fixture, skipif_known_failing

logging.basicConfig(level=logging.ERROR)

# Reset dependency graph before each test
# _safety_fixture, run_cell_ = make_safety_fixture(trace_messages_enabled=True)
_safety_fixture, run_cell_ = make_safety_fixture(enable_reactive_variables=True)


def run_cell(cell, **kwargs):
    # print()
    # print('*******************************************')
    # print('running', cell)
    # print('*******************************************')
    # print()
    run_cell_(cell, **kwargs)


def test_stack_on_tracing_reenable():
    run_cell(
        """
        x = 42

        def fake_func(y):
            z = 9
            return z

        new_xs = [fake_func(x) for _ in range(5)]

        logging.info(new_xs[0])
        """
    )


def test_non_idempotent_subscript():
    run_cell(
        """
        class IncDict:
            def __init__(self):
                self.val = 0
            def __getitem__(self, val):
                self.val += val
                return self.val
                
        d = IncDict()

        d_sub_1 = d[1]
        assert d_sub_1 == 1, f"got {d_sub_1}, but expected 1"
        """
    )


def test_starred_args():
    run_cell(
        """
        def f(foo, bar):
            return foo + bar
        """
    )
    run_cell("args = [1, 2]")
    run_cell("f(*args)")


def test_starred_assignment():
    run_cell("x = 0")
    run_cell("y = 1")
    run_cell("z = 2")
    run_cell('lst = ["foo", "bar"]')
    # just to make sure the tracer can handle a starred expr in list literal
    run_cell("s, *t = [x + 1, y + 2, z + 3, *lst]")
    run_cell("z = 42")
    run_cell("logging.info(s)")
    run_cell("logging.info(t[0])")
    run_cell("logging.info(t[1])")
    run_cell("x = 99")
    run_cell("logging.info(s)")
    run_cell("logging.info(t[0])")


def test_slices():
    run_cell("lst = list(range(10))")
    run_cell("foo = lst[3:7]")
    run_cell("lst[1:2] = foo")


def test_partial_slices():
    run_cell('s = "Hello, world!"')
    run_cell("logging.info(str(s[:7]))")
    run_cell("logging.info(str(s[7:]))")


def test_delete():
    run_cell("lst = list(range(10))")
    run_cell("del lst[0]")
    run_cell("del lst[-1]")


def test_delitem():
    run_cell(
        """
        class Foo:
            def __init__(self):
                self.lst = []
            def __getitem__(self, i):
                return self.lst[i]
            def __setitem__(self, i, v):
                self.lst[i] = v
            def __delitem__(self, i):
                del self.lst[i]
        """
    )
    run_cell("foo = Foo()")
    run_cell("foo.lst.append(0)")
    run_cell("foo.lst.append(1)")
    run_cell("foo.lst.append(2)")
    run_cell("logging.info(foo[0])")
    run_cell("foo[1] = 2")
    run_cell("del foo[-1]")


def test_empty_return():
    run_cell(
        """
        def foo():
            return
        """
    )
    run_cell("x = foo()")


def test_fancy_slices():
    run_cell(
        """
        class Foo:
            def __init__(self, x):
                self.x = x
        """
    )
    run_cell("foo = Foo(1)")
    run_cell("import numpy as np")
    run_cell("x = np.zeros((3, 3, 3))")
    run_cell("logging.info(x[:,:,:])")
    run_cell("logging.info(x[:,...])")
    run_cell("logging.info(x[1:,...])")
    run_cell("logging.info(x[foo.x])")
    run_cell("logging.info(x[foo.x:,...])")
    run_cell("logging.info(x[foo.x:foo.x+1,:,...])")


def test_fancy_slice_assign_augassign():
    run_cell("%safety trace_messages enable")
    run_cell("import numpy as np")
    run_cell("x = np.zeros((3, 3, 3))")
    run_cell("x[:, 1 ,...] = 1.")
    run_cell("x[:, 1 ,...] /= 5.")


def test_pass():
    run_cell("if True: pass")


def test_none_key():
    run_cell("d = {}")
    run_cell("d[None] = None")
    run_cell("d.clear()")


def test_global_var():
    run_cell("x = 0")
    run_cell("def f(): global x; x = 42")
    run_cell("f()")
    run_cell("assert x == 42")


def test_syntax_error_does_not_completely_mess_up_kernel():
    # first run a cell w/ syntax error, then one w/out
    # the second cell should be fine
    try:
        run_cell("x ++= 5")
    except SyntaxError:
        pass
    run_cell("x = 5")


def test_reactive_variable_does_not_break():
    run_cell("x = 0")
    run_cell("y = $x + 1")
    run_cell("print($y)")


def test_namespace_change_to_non_container():
    run_cell("x = 3, 4")
    run_cell("x = 3")
